from datetime import datetime, timezone
import os

from feedgen.feed import FeedGenerator

from flourish.generators.base import BaseGenerator
from flourish.generators.mixins import SourcesMixin


class AtomGenerator(SourcesMixin, BaseGenerator):
    order_by = ('-published')
    file_extension = '.atom'
    limit = 20
    required_config_keys = ('author', 'base_url', 'title')

    def get_objects(self, tokens):
        """
        Only consider objects that have the key "published" with a
        datetime value that is in the past.
        """
        _now = datetime.now().replace(tzinfo=timezone.utc)
        _sources = self.get_filtered_sources()
        _already_published = _sources.filter(published__lt=_now)
        _filtered = _already_published.filter(**tokens)
        _ordered = _filtered.order_by(self.order_by)

        if self.limit is not None:
            self.source_objects = _ordered[0:self.limit]
        else:
            self.source_objects = _ordered

        return self.source_objects

    def render_output(self):
        feed = FeedGenerator()
        feed.author({'name': self.flourish.site_config['author']})
        feed.title(self.flourish.site_config['title'])
        # feed.link(self.flourish.site_config['base_url'])
        feed.id('%s%s' % (
            self.flourish.site_config['base_url'],
            self.current_path,
        ))
        feed.link(href='%s%s' % (
                self.flourish.site_config['base_url'],
                self.current_path,
            ), rel='self')
        feed.link(href=self.flourish.site_config['base_url'], rel='alternate')

        last_updated = self.source_objects[0].published

        for _object in self.source_objects:
            entry = feed.add_entry(order='append')
            entry.title(self.get_entry_title(_object))
            entry.id(self.get_entry_id(_object))
            entry.link(href=_object.absolute_url, rel='alternate')
            entry.published(_object.published)
            entry.content(
                content=self.get_entry_content(_object),
                type='html'
            )

            if 'author' in _object:
                entry.author({'name': _object.author})
            else:
                entry.author({'name': self.flourish.site_config['author']})

            if 'updated' in _object:
                entry.updated(_object.updated)
                if _object.updated > last_updated:
                    last_updated = _object.updated
            else:
                entry.updated(_object.published)

        feed.updated(last_updated)

        return feed.atom_str(pretty=True)

    def get_entry_content(self, object):
        return object.body

    def get_entry_title(self, object):
        return object.title

    def get_entry_id(self, object):
        return object.absolute_url

    # FIXME refactor
    def output_to_file(self):
        _filename = self.get_output_filename()
        if self.report:
            print('->', _filename)

        _rendered = self.render_output()
        _directory = os.path.dirname(_filename)
        if not os.path.isdir(_directory):
            os.makedirs(_directory)
        with open(_filename, 'wb') as _output:
            _output.write(_rendered)



# encoding: utf8

from collections import defaultdict
from datetime import date, datetime
from importlib.machinery import SourceFileLoader
from operator import itemgetter
import os
from shutil import copyfile, rmtree
import sys
import warnings

from jinja2 import Environment, FileSystemLoader
import toml

from .lib import relative_list_of_files_in_directory
from .sectileloader import SectileLoader
from .source import (
    JsonSourceFile,
    MarkdownSourceFile,
    SourceFile,
    TomlSourceFile,
    CsvSourceFile,
)
from .sourcelist import SourceList
from .version import __version__    # noqa: F401

sys.dont_write_bytecode = True


class Flourish(object):
    def __init__(
        self,
        source_dir='source',
        templates_dir='templates',
        output_dir='output',
        sass_dir='sass',
        fragments_dir=None,
        future=None,
        skip_scan=False,
    ):
        self.source_dir = source_dir
        self.templates_dir = templates_dir
        self.fragments_dir = fragments_dir
        self.output_dir = output_dir
        self.sass_dir = sass_dir
        self.future = future
        self._assets = {}
        self._cache = {}
        self._source_files = []
        self._source_path = None
        self._paths = {}

        # using sectile fragments overrides standard filesystem templates
        if self.fragments_dir:
            self.using_sectile = True
            self.jinja = Environment(
                loader=SectileLoader(self.fragments_dir),
                keep_trailing_newline=True,
            )
        else:
            self.using_sectile = False
            self.jinja = Environment(
                loader=FileSystemLoader(self.templates_dir),
                keep_trailing_newline=True,
            )

        self.jinja.globals['path'] = self.template_resolve_path
        self.jinja.globals['lookup'] = self.template_get

        if not os.path.isdir(self.source_dir):
            raise Flourish.RuntimeError(
                'The source directory "%s" must exist' % self.source_dir)

        self.site_config = self._read_site_config()
        if not skip_scan:
            self._rescan_sources()

        try:
            generate = SourceFileLoader(
                    'generate', '%s/generate.py' % self.source_dir
                ).load_module()
        except FileNotFoundError:
            raise Flourish.RuntimeError(
                ('The source directory "%s"'
                 ' must contain the file "generate.py"')
                    % self.source_dir
            )

        try:
            self.set_global_context(getattr(generate, 'GLOBAL_CONTEXT'))
        except AttributeError:
            # global context is optional
            pass

        try:
            self.add_template_filters(getattr(generate, 'TEMPLATE_FILTERS'))
        except AttributeError:
            # filters are optional
            pass

        try:
            for path in generate.PATHS:
                self.add_path(path)
        except AttributeError:
            raise Flourish.RuntimeError(
                'There are no paths configured in generate.py')

    @property
    def sources(self):
        try:
            future = self.site_config['future'] 
        except KeyError:
            future = True

        # object instantiation takes precedence over site config
        if self.future is not None:
            future = self.future

        return SourceList(
            self._source_files,
            future = future
        )

    def get(self, slug):
        """ Get a single source document by slug. """
        try:
            return self._cache[slug]
        except KeyError:
            raise SourceFile.DoesNotExist

    def template_get(self, slug):
        try:
            return self.get(slug)
        except SourceFile.DoesNotExist:
            warnings.warn('Cannot get "%s" as it does not exist' % slug)
            return None

    def add_path(self, path):
        path.setup(self)
        if path.name == 'source':
            self._source_path = path
        if hasattr(path, 'required_config_keys'):
            for key in path.required_config_keys:
                if key not in self.site_config:
                    raise Flourish.MissingKey(
                        '%s requires an entry "%s" in _site.toml' % (
                            path.__class__.__name__,
                            key,
                        ))
        self._paths.update({path.name: path})

    def resolve_path(self, name, **kwargs):
        return self._paths[name].resolve(**kwargs)

    def template_resolve_path(self, name, **kwargs):
        try:
            return self.resolve_path(name, **kwargs)
        except KeyError:
            warnings.warn(
                'Cannot resolve path "%s", it has missing arguments' % name
            )
            return ''

    def get_handler_for_path(self, path):
        matches = []
        for key in self._paths:
            _dicts = self._paths[key].can_generate(path)
            if _dicts:
                for _dict in _dicts:
                    matches.append((key, _dict))
        return matches

    def all_valid_filters_for_path(self, name):
        if name in self._paths:
            return self._paths[name].all_valid_filters()
        else:
            raise SourceFile.DoesNotExist

    def generate_site(self, report=False):
        if os.path.exists(self.output_dir):
            rmtree(self.output_dir)
        os.makedirs(self.output_dir)
        self.generate_all_paths(report=report)
        self.copy_assets(report=report)

    def generate_all_paths(self, report=False):
        for path in self._paths:
            self._paths[path].generate(report)

    def generate_path(self, path, report=False):
        handlers = self.get_handler_for_path(path)
        for key, args in handlers:
            path = self._paths[key]
            path.generate(report, tokens=[args])

    def path_recipe(self, path):
        handlers = self.get_handler_for_path(path)
        for key, args in handlers:
            path = self._paths[key]
            recipe = path.get_recipe(tokens=[args])
            if recipe:
                return recipe
        return None

    def set_global_context(self, global_context):
        self.global_context = global_context

    def add_template_filters(self, filters):
        for key, value in filters.items():
            self.jinja.filters[key] = value

    def copy_assets(self, report=False):
        for _file in self._assets:
            _source = '%s/%s' % (self.source_dir, _file)
            _destination = '%s/%s' % (self.output_dir, _file)
            _directory = os.path.dirname(_destination)
            if not os.path.isdir(_directory):
                os.makedirs(_directory)
            copyfile(_source, _destination)
            if report:
                print('++', _destination)

    def _read_site_config(self):
        _config_file = '%s/_site.toml' % self.source_dir
        try:
            with open(_config_file) as _file:
                return toml.loads(_file.read())
        except FileNotFoundError:
            return {}

    def _rescan_sources(self):
        """ Find source documents and register them. """
        _seen = {}
        for _file in relative_list_of_files_in_directory(self.source_dir):
            if _file == '_site.toml':
                continue
            if _file.startswith('generate.py'):
                continue
            slug, _ = os.path.splitext(_file)
            timestamp = os.stat(os.path.join(self.source_dir, _file)).st_mtime
            try:
                source = self.get(slug)
                assert timestamp == source.timestamp
                _seen[slug] = 1
            except:     # noqa: E722
                # FIXME check for slug-ishness and otherwise ignore
                # (this could simplify _site.toml by being just another
                # ignored filename?)
                is_attachment_file = (
                    _file.endswith(('.markdown', '.html')) and
                    len(_file.split('.')) == 3
                )

                if _file.endswith('.toml'):
                    self._cache[slug] = TomlSourceFile(self, _file)
                    _seen[slug] = 1
                elif (
                    _file.endswith('.markdown')
                    and len(_file.split('.')) == 2
                ):
                    self._cache[slug] = MarkdownSourceFile(self, _file)
                    _seen[slug] = 1
                elif _file.endswith('.json'):
                    self._cache[slug] = JsonSourceFile(self, _file)
                    _seen[slug] = 1
                elif _file.endswith('.csv'):
                    for src in CsvSourceFile(self, _file).get_sources():
                        if src['slug'] in _seen:
                            warnings.warn(
                                (
                                    'Existing source "%s" has been '
                                    'overriden by "%s"'
                                ) % (
                                    src['slug'],
                                    src,
                                )
                            )
                        self._cache[src['slug']] = src
                        _seen[src['slug']] = 1
                elif not is_attachment_file:
                    self._assets[_file] = True

        # remove anything no longer there
        # FIXME will also need output removing
        _removed = {}
        for source in self._cache:
            if source not in _seen:
                _removed[source] = 1
        for source in _removed:
            del self._cache[source]

        self._source_files = self._cache.values()

    @property
    def publication_dates(self):
        def recursively_default_dict():
            return defaultdict(recursively_default_dict)

        _captured = recursively_default_dict()
        for _source in self.sources.all():
            try:
                _date = getattr(_source, 'published')
                _captured[_date.year][_date.month][_date.day] = 1
            except AttributeError:
                pass

        _dates = []
        for _year in sorted(_captured):
            _months = []
            for _month in sorted(_captured[_year]):
                _days = []
                for _day in sorted(_captured[_year][_month]):
                    _days.append(date(_year, _month, _day))
                _months.append({
                    'month': date(_year, _month, 1),
                    'days': _days
                })
            _dates.append({'year': date(_year, 1, 1), 'months': _months})
        return _dates

    def get_valid_filters_for_tokens(self, tokens, objects=None):
        if objects is None:
            objects = self.sources

        _first_token = tokens[0]

        _values = set()
        for _source in objects:
            _is_date_filter = (
                _first_token in ['year', 'month', 'day'] and
                'published' in _source and
                type(_source['published'] == datetime)
            )

            if _is_date_filter:
                _value = getattr(_source['published'], _first_token)
                _values.add('%02d' % _value)
            elif _first_token in _source:
                if type(_source[_first_token]) == list:
                    for _value in _source[_first_token]:
                        _values.add(_value)
                else:
                    _values.add(_source[_first_token])

        _filters = []
        for _value in _values:
            _dict = {_first_token: _value}
            if len(tokens) == 1:
                _filters.append(_dict)
            else:
                _sub_objects = objects.filter(**_dict)
                _sub_tokens = self.get_valid_filters_for_tokens(
                    tokens[1:], _sub_objects)
                for _sub_token in _sub_tokens:
                    _update = _dict.copy()
                    _update.update(_sub_token)
                    _filters.append(_update)

        # the sort is only so the tests can compare results easily
        return sorted(_filters, key=itemgetter(_first_token))

    def __repr__(self):
        return '<flourish.Flourish object (source=%s)>' % self.source_dir


    class MissingKey(Exception):
        pass

    class RuntimeError(Exception):
        pass




import json

from jinja2 import BaseLoader, TemplateNotFound
from sectile import Sectile


class SectileLoader(BaseLoader):
    def __init__(self, fragments_dir):
        self.sectile = Sectile(fragments=fragments_dir)
        self.prepared_path = None
        self.prepared_base = None
        self.prepared_dimensions = None
        self.cache = {}

    def dimensions(self):
        return self.sectile.get_dimensions_list()

    def generate_template(self, path, base_template, **dimensions):
        fingerprint = '%s-%s-%s' % (
            path,
            base_template,
            json.dumps(dimensions)
        )
        content, fragments = self.sectile.generate(
            path,
            base_template,
            **dimensions
        )
        self.cache = {
            'path': path,
            'fingerprint': fingerprint,
            'base_template': base_template,
            'dimensions': dimensions,
            'content': content,
            'fragments': fragments,
        }
        return self.cache

    def prepare_template(self, path, base_template, **dimensions):
        generated = self.generate_template(path, base_template, **dimensions)
        return generated['fingerprint']

    def get_source(self, environment, fingerprint):
        if self.cache['fingerprint'] != fingerprint:
            raise TemplateNotFound(
                "%s, %s {%s}" % (
                    self.prepared_path,
                    self.prepared_base,
                    self.prepared_dimensions
                )
            )
        else:
            return self.cache['content'], None, None



# encoding: utf8


def publication_range(flourish, key='published'):
    _lowest = None
    _highest = None

    for _source in flourish.sources.all():
        try:
            _pub = getattr(_source, key)
            _year = _pub.year
            if _lowest is None or _year < _lowest:
                _lowest = _year
            if _highest is None or _year > _highest:
                _highest = _year
        except AttributeError:
            pass

    if _lowest is not None and _highest is not None:
        return u'%d–%d' % (_lowest, _highest)



import os

import sass

from flourish.generators.base import BaseGenerator


# FIXME
# when refreshing sources, also refresh generators
class SassGenerator(BaseGenerator):
    output_style = 'expanded'
    file_extension = '.css'
    sass_sources = []

    def setup(self, flourish):
        super().setup(flourish)
        self.find_sass_sources()

    def find_sass_sources(self):
        for root, dirs, files in os.walk(self.flourish.sass_dir):
            root = root[len(self.flourish.sass_dir):]
            for file in files:
                base, ext = os.path.splitext(file)
                if not base.startswith('_') and ext == '.scss':
                    self.sass_sources.append(os.path.join(root, base))

    def get_path_tokens(self):
        tokens = []
        for source in self.sass_sources:
            tokens.append({'sass_source': source})
        return tokens

    def all_valid_filters(self):
        valid_filters = []
        args = self.arguments

        if len(args) == 0:
            valid_filters.append({})
        elif args == ['sass_source']:
            for source in self.sass_sources:
                valid_filters.append({'sass_source': source})
        else:
            raise ValueError
        return valid_filters

    def get_objects(self, tokens):
        # there are no Source objects
        pass

    def render_output(self):
        source = os.path.join(
            self.flourish.sass_dir,
            '%s.scss' % self.tokens['sass_source'],
        )
        return sass.compile(
            filename=source,
            output_style=self.output_style,
        )



import argparse
from datetime import datetime
from hashlib import md5
import mimetypes
import os
import sys
from textwrap import dedent
import time

import boto3
from flask import Flask, request, send_from_directory
from jinja2 import Template

from . import Flourish, __version__
from .examples import example_files
from .lib import relative_list_of_files_in_directory


recipe_html = Template(dedent("""\
    <html>
    <head>
      <style>
        #dimensions dt {
          float: left;
          margin-right: 20px;
          min-width: 100px;
        }
        #dimensions dd {
          font-weight: bold;
        }
        #template ol {
          margin: 0;
          padding: 0;
        }
        #template ol ol li {
          padding: 0 0 0 17px;
          border-left: 1px solid #999;
          margin-left: 3px;
          margin-bottom: 15px;
        }
        #template li {
          list-style: none;
          margin-top: 10px;
        }
        #template li i {
          color: #66a;
          font-size: 0.8em;
        }
        #template li.missing {
          margin-bottom: 10px;
        }
        #template li.missing > i {
          font-weight: bold;
          color: #f99;
        }
        #template li pre {
          margin: 0;
        }
      </style>
    </head>
    <body>
      <h1>Recipe for {{path|e}}</h1>
      <div id='template'>
        {% if sectile_dimensions %}
          <h2>Dimensions</h2>
          <dl id='dimensions'>
          {% for dim in sectile_dimensions %}
            <dt>{{dim}}</dt>
            <dd>{{sectile_dimensions[dim]}}</dd>
          {% endfor %}
          </dl>
        {% endif %}
        <h2>Template: <i>{{template_name|e}}</i></h2>
        {% if sectile_fragments %}
          {% set depth = namespace(previous=-1) %}
          {% for fragment in sectile_fragments %}
            {% if depth.previous < fragment.depth %}
              <ol>
            {% elif depth.previous > fragment.depth %}
              {% for i in range(fragment.depth, depth.previous) %}
                </li></ol>
              {% endfor %}
            {% else %}
              </li>
            {% endif %}
            <li class='{% if not fragment.found %}missing{% endif %}'>
            {% if not fragment.found %}
              <i>{{fragment.file}}: None</i>
            {% else %}
              <i>{{fragment.found}}</i>
              <pre>{{fragment.fragment|e}}</pre>
            {% endif %}

            {% if depth.previous < fragment.depth %}

            {% elif depth.previous > fragment.depth %}
            {% else %}
              </li>
            {% endif %}
            {% set depth.previous = fragment.depth %}
          {% endfor %}
          </ol>
          {% for i in range(depth.previous) %}
            </li></ol>
          {% endfor %}
          <p>Result:</p>
        {% endif %}
        <pre>{{template|e}}</pre>
      </div>
      <h2>Context:</h2>
      <div id='context'>{{context|e}}</div>
    </body>
    </html>
    """))


def main():
    version = ('Flourish static site generator, version %s -- '
               'http://withaflourish.net' % __version__)
    parser = argparse.ArgumentParser(
        description=(
            'Flourish generates websites from source data and templates.'
        ),
        epilog=(
            'Version %s -- http://withaflourish.net' % __version__
        ),
    )
    parser.add_argument(
        '--base',
        default=None,
        help='Base directory that contains all others (default: .)',
    )
    parser.add_argument(
        '--source',
        default='source',
        help='Directory containing source files (default: %(default)s)',
    )
    parser.add_argument(
        '--templates',
        default='templates',
        help='Directory containing templates (default: %(default)s)',
    )
    parser.add_argument(
        '--fragments',
        default='fragments',
        help=(
            'Directory containing Sectile template fragments'
            ' (experimental feature, no default)'
        ),
    )
    parser.add_argument(
        '--output',
        default='output',
        help='Directory to output to (default: %(default)s)',
    )
    parser.add_argument(
        '--version',
        action='store_true',
        help='Report the version of flourish',
    )
    subparsers = parser.add_subparsers(
        title='Actions',
        description=(
            'To get more information, run "flourish [command] --help".'
        ),
        dest='action',
    )

    parser_generate = subparsers.add_parser(
        'generate',
        help='creates the output HTML',
    )
    parser_generate.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='Report each URL as it is generated'
    )
    parser_generate.add_argument(
        '--include-future',
        action='store_true',
        help=(
            'Include sources with a publication date in the future, '
            'overriding the "future" setting in _site.toml.'
        ),
    )
    parser_generate.add_argument(
        '--exclude-future',
        action='store_true',
        help=(
            'Exclude sources with a publication date in the future, '
            'overriding the "future" setting in _site.toml.'
        ),
    )
    parser_generate.add_argument(
        'path',
        nargs='*',
        help=(
            'Generate only this path (appending ? makes it a wildcard '
            'match to generate anything that starts with this path '
            ' -- eg. /2020/?). Can be specified repeatedly.'
        ),
    )

    parser_preview = subparsers.add_parser(
        'preview',
        help=(
            'preview the generated content from your local computer '
            'before uploading'
        ),
    )
    parser_preview.add_argument(
        '-p', '--port',
        default='3567',
        help='Port on which to bind preview webserver (default: %(default)s)',
    )
    parser_preview.add_argument(
        '-g', '--generate',
        action='store_true',
        help='Regenerate the requested page before serving',
    )

    parser_upload = subparsers.add_parser(
        'upload',
        help='upload the generated site to an AWS S3 bucket',
    )
    parser_upload.add_argument(
        '--dry-run',
        action='store_true',
        help='only list what would be uploaded',
    )
    parser_upload.add_argument(
        '--invalidate',
        action='store_true',
        help='invalidate uploaded paths in CloudFront',
    )
    parser_upload.add_argument(
        'bucket',
        nargs='?',
        help='bucket to upload to (default: bucket entry in _site.toml)',
    )
    parser_upload.add_argument(
        'cloudfront_id',
        nargs='?',
        help='the CloudFront distribution ID',
    )

    parser_example = subparsers.add_parser(             # noqa: F841
        'example',
        help=(
            'creates a new flourish website in the '
            'current directory with some example content'
        ),
    )

    args = parser.parse_args()
    if args.version:
        print(version)
    elif args.action is None:
        parser.print_help()
    else:
        if args.base is not None:
            args.source = os.path.join(args.base, args.source)
            args.templates = os.path.join(args.base, args.templates)
            args.fragments = os.path.join(args.base, args.fragments)
            args.output = os.path.join(args.base, args.output)
        # find default 'fragments' directory but without forcing it
        if not os.path.isdir(args.fragments):
            args.fragments = None
        action = ACTIONS[args.action]
        action(args)


def generate(args):
    future = None
    if args.exclude_future:
        future = False
    if args.include_future:
        future = True
    try:
        flourish = Flourish(
            source_dir=args.source,
            templates_dir=args.templates,
            fragments_dir=args.fragments,
            output_dir=args.output,
            future=future,
        )
    except Flourish.MissingKey as e:
        sys.exit('Error: %s' % str(e))
    except Flourish.RuntimeError as e:
        sys.exit('Error: %s' % str(e))
    if args.path:
        for path in args.path:
            flourish.generate_path(path, report=args.verbose)
    else:
        flourish.generate_site(report=args.verbose)


def preview_server(args):
    output_dir = os.path.abspath(args.output)
    try:
        flourish = Flourish(
            source_dir=args.source,
            templates_dir=args.templates,
            fragments_dir=args.fragments,
            output_dir=args.output,
        )
    except Flourish.MissingKey as e:
        sys.exit('Error: %s' % str(e))
    except Flourish.RuntimeError as e:
        sys.exit('Error: %s' % str(e))
    app = Flask(__name__)

    # TODO 3xx redirects
    # TODO 404 page
    # TODO 410 page
    @app.route('/')
    @app.route('/<path:path>')
    def send_file(path=''):
        generate = '/%s' % path

        if 'showrecipe' in request.args:
            return recipe_html.render(flourish.path_recipe(generate))
        else:
            # fix URLs for .html
            filename = os.path.basename(path)
            if filename == '':
                path = '%sindex.html' % path
            _, ext = os.path.splitext(path)
            if ext == '':
                path += '.html'

            # regenerate if requested
            if args.generate:
                flourish._rescan_sources()
                flourish.generate_path(generate, report=True)

            return send_from_directory(output_dir, path)

    @app.after_request
    def add_header(response):
        response.cache_control.max_age = 0
        return response

    app.run(port=args.port)


def create_example(args):
    for _dir in ['source', 'templates']:
        if not os.path.isdir(_dir):
            os.mkdir(_dir)
    for _filename in example_files:
        with open(_filename, 'w') as _example_file:
            _example_file.write(example_files[_filename])
    generate(argparse.Namespace(
        source='source',
        templates='templates',
        output='output',
        action='generate',
        verbose=True,
        path=[],
    ))
    print('Example site created: run "flourish preview --generate"')
    print('and go to http://localhost:3567/ in your browser to see the site.')
    print('Then click on "Welcome to your new blog" to get started.')


def upload(args):
    mimetypes.init()

    try:
        flourish = Flourish(
            source_dir=args.source,
            templates_dir=args.templates,
            fragments_dir=args.fragments,
            output_dir=args.output,
            skip_scan=True,
        )
    except Flourish.MissingKey as e:
        sys.exit('Error: %s' % str(e))
    except Flourish.RuntimeError as e:
        sys.exit('Error: %s' % str(e))

    _bucket_name = args.bucket
    _cloudfront = args.cloudfront_id
    if _bucket_name is None:
        _bucket_name = flourish.site_config['bucket']
    if args.invalidate and args.cloudfront_id is None:
        _cloudfront = flourish.site_config['cloudfront_id']

    # FIXME remove nonexistent files
    _s3 = boto3.resource('s3')
    _bucket = _s3.Bucket(_bucket_name)
    _files = relative_list_of_files_in_directory(args.output)
    _objects = dict()
    _invalidations = []
    for _object in _bucket.objects.all():
        _objects.update({_object.key: _object})

    for _path in _files:
        _file = os.path.basename(_path)
        _type, _ = mimetypes.guess_type(_file)
        _s3path = _path
        if _file.endswith('.html'):
            _type = 'text/html'
            if not _file == 'index.html':
                _s3path = _path[:-5]

        _upload = True
        _md5 = '"%s"' % md5(
                open('%s/%s' % (args.output, _path), 'rb').read()
            ).hexdigest()
        try:
            if _md5 == _objects[_s3path].e_tag:
                _upload = False
        except KeyError:
            pass

        if _upload:
            print('>>', _s3path)
            if not args.dry_run:
                _object_args = {
                    'Key': _s3path,
                    'ACL': 'public-read',
                    'Body': open('%s/%s' % (args.output, _path), 'rb'),
                    'ContentType': _type,
                }
                _bucket.put_object(**_object_args)
                invalidate = '/' + _s3path
                if invalidate.endswith('index.html'):
                    invalidate = invalidate[:-10]
                _invalidations.append(invalidate)

    if args.invalidate and _invalidations:
        cf = boto3.client('cloudfront')
        result = cf.create_invalidation(
            DistributionId=_cloudfront,
            InvalidationBatch={
                'Paths': {
                    'Quantity': len(_invalidations),
                    'Items': _invalidations,
                },
                'CallerReference': datetime.now().isoformat(),
            }
        )
        invalidation_id = result['Invalidation']['Id']
        print('Invalidating distribution %s: %s' % (_cloudfront, invalidation_id))
        while True:
            time.sleep(15)
            status = cf.get_invalidation(
                DistributionId = _cloudfront,
                Id = invalidation_id,
            )
            print(
                datetime.now().strftime('%H:%M:%S'),
                invalidation_id,
                status['Invalidation']['Status'],
            )
            if status['Invalidation']['Status'] == 'Completed':
                break

ACTIONS = {
    'generate': generate,
    'preview': preview_server,
    'example': create_example,
    'upload': upload,
}

if __name__ == '__main__':
    main()
